package com.linbit.linstor.layer.drbd.utils;

import com.linbit.ChildProcessTimeoutException;
import com.linbit.extproc.ExtCmd;
import com.linbit.extproc.ExtCmd.OutputData;
import com.linbit.extproc.ExtCmdFactory;
import com.linbit.extproc.ExtCmdFailedException;
import com.linbit.linstor.core.identifier.VolumeNumber;
import com.linbit.linstor.core.objects.Resource;
import com.linbit.linstor.core.types.MinorNumber;
import com.linbit.linstor.core.types.NodeId;
import com.linbit.linstor.layer.storage.utils.Commands;
import com.linbit.linstor.layer.storage.utils.Commands.RetryHandler;
import com.linbit.linstor.storage.StorageException;
import com.linbit.linstor.storage.data.adapter.drbd.DrbdRscData;
import com.linbit.linstor.storage.data.adapter.drbd.DrbdVlmData;

import javax.inject.Inject;
import javax.inject.Singleton;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

@Singleton
public class DrbdAdm
{
    public static final String DRBDADM_UTIL   = "drbdadm";
    public static final String DRBDMETA_UTIL  = "drbdmeta";
    public static final String DRBDSETUP_UTIL = "drbdsetup";

    public static final String ALL_KEYWORD = "all";
    public static final String DRBDCTRL_RES_NAME = ".drbdctrl";

    public static final int DRBDMETA_NO_VALID_MD_RC = 255;
    public static final int DRBDMETA_STRANGE_BM_OFFSET = 1;

    public static final int WAIT_CONNECT_RES_TIME = 10;

    private final ExtCmdFactory extCmdFactory;

    @Inject
    public DrbdAdm(ExtCmdFactory extCmdFactoryRef)
    {
        extCmdFactory = extCmdFactoryRef;
    }

    /**
     * Adjusts a resource
     */
    public void adjust(
        DrbdRscData<Resource> drbdRscData,
        boolean skipNet,
        boolean skipDisk,
        boolean discard
    )
        throws ExtCmdFailedException
    {
        List<String> command = new ArrayList<>();
        command.addAll(Arrays.asList(DRBDADM_UTIL, "-vvv", "adjust"));

        // Using -c disables /etc/drbd.d/global_common.conf
        // command.add("-c");
        // command.add("/etc/drbd.d/" + resourceName.displayValue + ".res");

        if (discard)
        {
            command.add("--discard-my-data");
        }
        if (skipNet)
        {
            command.add("--skip-net");
        }
        if (skipDisk)
        {
            command.add("--skip-disk");
        }
        // String resName = resourceName.value;
        // command.addAll(asConfigParameter(resName));
        // if (volNum != null)
        // {
        //     resName += "/" + volNum.value;
        // }
        // command.add(resName);
        command.add(drbdRscData.getSuffixedResourceName());
        // execute(Arrays.asList("drbdsetup", "show", drbdRscData.getSuffixedResourceName()));
        execute(command);
        // execute(Arrays.asList("drbdsetup", "show", drbdRscData.getSuffixedResourceName()));
    }

    /**
     * Resizes a resource
     */
    public void resize(
        DrbdVlmData<Resource> drbdVlmData,
        boolean assumeClean,
        Long shrinkToSize
    )
        throws ExtCmdFailedException
    {
        DrbdRscData<Resource> drbdRscData = drbdVlmData.getRscLayerObject();
        waitConnectResource(drbdRscData, WAIT_CONNECT_RES_TIME);
        List<String> command = new ArrayList<>();
        command.addAll(Arrays.asList(DRBDADM_UTIL, "-vvv"));
        if (assumeClean)
        {
            command.add("--");
            command.add("--assume-clean");
        }
        if (shrinkToSize != null)
        {
            command.add("--size=" + Long.toString(shrinkToSize) + "K");
        }
        String resName = drbdRscData.getSuffixedResourceName();
        // Using -c disables /etc/drbd.d/global_common.conf
        // command.addAll(asConfigParameter(resName)); // basically just adds the -c <rscName.conf> as parameter
        command.add("resize");
        command.add(resName + "/" + drbdVlmData.getVlmNr().value);
        execute(command);
    }

    /**
     * Shuts down (unconfigures) a DRBD resource
     * @param rscNameSuffix
     */
    public void down(DrbdRscData<Resource> drbdRscData) throws ExtCmdFailedException
    {
        simpleSetupCommand(drbdRscData, (VolumeNumber) null, "down");
    }

    /**
     * Switches a DRBD resource to primary mode
     *
     * @throws StorageException
     */
    public void primary(
        DrbdRscData<Resource> drbdRscData,
        boolean force,
        boolean withDrbdSetup
    )
        throws StorageException
    {

        List<String> command = new ArrayList<>();
        if (withDrbdSetup)
        {
            command.add(DRBDSETUP_UTIL);
            if (force)
            {
                command.add("--force");
            }
            command.add("primary");
            command.add(drbdRscData.getSuffixedResourceName());
        }
        else
        {
            command.add(DRBDADM_UTIL);
            command.add("-vvv");
            command.add("primary");
            if (force)
            {
                command.add("--force");
            }
            command.add(drbdRscData.getSuffixedResourceName());
        }


        String[] commandArr = command.toArray(new String[command.size()]);
        Commands.genericExecutor(
            extCmdFactory.create(),
            commandArr,
            "Failed to set resource '" + drbdRscData.getSuffixedResourceName() + "' to primary",
            "Failed to set resource '" + drbdRscData.getSuffixedResourceName() + "' to primary",
            new RetryHandler()
            {
                private int retryCount = 3;

                @Override
                public boolean skip(OutputData outDataRef)
                {
                    return false;
                }

                @Override
                public boolean retry(OutputData outputDataRef)
                {
                    String errStr = new String(outputDataRef.stderrData);
                    boolean retryFlag = errStr.contains("Concurrent state changes detected and aborted") &&
                        retryCount > 0;
                    if (retryFlag)
                    {
                        --retryCount;
                        try
                        {
                            Thread.sleep(500);
                        }
                        catch (InterruptedException ignored)
                        {
                        }
                    }
                    return retryFlag;
                }
            }
        );
    }

    /**
     * Switches a resource to secondary mode
     */
    public void secondary(DrbdRscData<Resource> drbdRscData) throws ExtCmdFailedException
    {
        simpleAdmCommand(drbdRscData, "secondary");
    }

    /**
     * Connects a resource to its peer resuorces on other hosts
     */
    public void connect(DrbdRscData<Resource> drbdRscData, boolean discard) throws ExtCmdFailedException
    {
        adjust(drbdRscData, false, true, discard);
    }

    /**
     * Disconnects a resource from its peer resources on other hosts
     */
    public void disconnect(DrbdRscData<Resource> drbdRscData) throws ExtCmdFailedException
    {
        simpleAdmCommand(drbdRscData, "disconnect");
    }

    /**
     * Attaches a volume to its disk
     *
     * @throws ExtCmdFailedException
     */
    public void attach(DrbdVlmData<Resource> drbdVlmData) throws ExtCmdFailedException
    {
        adjust(drbdVlmData.getRscLayerObject(), true, false, false);
    }

    /**
     * Detaches a volume from its disk
     *
     * @param diskless If true, convert to a diskless client volume
     */
    public void detach(DrbdVlmData<Resource> drbdVlmData, boolean diskless) throws ExtCmdFailedException
    {
        List<String> commands = new ArrayList<>();
        commands.add(DRBDSETUP_UTIL);
        commands.add("detach");
        commands.add(Integer.toString(drbdVlmData.getVlmDfnLayerObject().getMinorNr().value));

        if (diskless)
        {
            commands.add("--diskless");
        }
        execute(commands);
    }

    /**
     * Calls drbdadm to create the metadata information for a volume
     */
    public void createMd(DrbdVlmData<Resource> drbdVlmData, int peers) throws ExtCmdFailedException
    {
        simpleAdmCommand(
            drbdVlmData.getRscLayerObject(),
            drbdVlmData.getVlmNr(),
            "--max-peers", Integer.toString(peers),
            "--",
            "--force",
            "create-md"
        );
    }

    /**
     * Calls drbdmeta to determine whether DRBD meta data exists on a volume
     */
    public boolean hasMetaData(String blockDevPath, int minorNr, String mdTypeParam) throws ExtCmdFailedException
    {
        // It is probably safer to fail because missing meta data was not detected than
        // to overwrite existing meta data because it was not detected, therefore default
        // to true, indicating that meta data exists
        boolean mdFlag = true;

        String[] params = getGetGiCommand(minorNr, blockDevPath, mdTypeParam);
        ExtCmd utilsCmd = extCmdFactory.create();
        File nullDevice = new File("/dev/null");
        try
        {
            OutputData outputData = utilsCmd.pipeExec(ProcessBuilder.Redirect.from(nullDevice), params);
            if (outputData.exitCode == DRBDMETA_NO_VALID_MD_RC || outputData.exitCode == DRBDMETA_STRANGE_BM_OFFSET)
            {
                mdFlag = false;
            }
        }
        catch (ChildProcessTimeoutException timeoutExc)
        {
            throw new ExtCmdFailedException(params, timeoutExc);
        }
        catch (IOException ioExc)
        {
            throw new ExtCmdFailedException(params, ioExc);
        }
        return mdFlag;
    }

    public String getCurrentGID(String blockDevPath, int minorNr, String mdTypeParam) throws ExtCmdFailedException
    {
        String ret;

        String[] params = getGetGiCommand(minorNr, blockDevPath, mdTypeParam);
        ExtCmd utilsCmd = extCmdFactory.create();
        try
        {
            OutputData outputData = utilsCmd.exec(params);
            ret = new String(outputData.stdoutData);
        }
        catch (ChildProcessTimeoutException timeoutExc)
        {
            throw new ExtCmdFailedException(params, timeoutExc);
        }
        catch (IOException ioExc)
        {
            throw new ExtCmdFailedException(params, ioExc);
        }
        return ret;
    }

    private static String[] getGetGiCommand(int minorNr, String blockDevPath, String mdTypeParam)
    {
        return new String[] {
            DRBDMETA_UTIL,
            Integer.toString(minorNr),
            "v09",
            blockDevPath,
            mdTypeParam,
            "get-gi",
            "--node-id",
            "0",
            "--force"// force is needed if the drbd-resource is still in Negotiating state or earlier.
            // in that case, drbdmeta asks "Exclusive open failed. Do it anyways?" and expects to type 'yes'.
            // should not break anything
        };
    }

    /**
     * Calls drbdmeta to create the metadata information for a volume
     */
    public void setGi(
        NodeId nodeId,
        MinorNumber minorNr,
        String blockDevPath,
        String currentGi,
        String history1Gi,
        boolean upToDateData,
        boolean internal
    )
        throws ExtCmdFailedException
    {
        String giData = currentGi + ":";
        if (upToDateData || history1Gi != null)
        {
            String checkedHistory1Gi = history1Gi == null ? "0" : history1Gi;
            giData += "0:" + checkedHistory1Gi + ":0:";
            if (upToDateData)
            {
                giData += "1:1:";
            }
        }
        execute(
            DRBDMETA_UTIL,
            "--force",
            "--node-id", Integer.toString(nodeId.value),
            String.valueOf(minorNr.value),
            "v09",
            blockDevPath,
            internal ? "internal" : "flex-external",
            "set-gi", giData
        );
    }

    /**
     * For debug purposes
     *
     * @param vlmDataRef
     * @throws ExtCmdFailedException
     */
    public void showGi(DrbdVlmData<Resource> vlmDataRef) throws ExtCmdFailedException
    {
        DrbdRscData<Resource> rscData = vlmDataRef.getRscLayerObject();
        execute(
            DRBDSETUP_UTIL,
            "show-gi",
            rscData.getSuffixedResourceName(),
            rscData.getNodeId().value + "",
            vlmDataRef.getVlmNr().value + ""
        );
    }

    public void suspendIo(DrbdRscData<Resource> drbdRscData)
        throws ExtCmdFailedException
    {
        execute(DRBDADM_UTIL, "suspend-io", drbdRscData.getSuffixedResourceName());
    }

    public void resumeIo(DrbdRscData<Resource> drbdRscData)
        throws ExtCmdFailedException
    {
        if (drbdRscData.resFileExists())
        {
            execute(DRBDADM_UTIL, "resume-io", drbdRscData.getSuffixedResourceName());
        }
        else
        {
            for (DrbdVlmData<Resource> drbdVlmData : drbdRscData.getVlmLayerObjects().values())
            {
                execute(
                    DRBDSETUP_UTIL,
                    "resume-io",
                    Integer.toString(drbdVlmData.getVlmDfnLayerObject().getMinorNr().value)
                );
            }
        }
    }

    public void waitConnectResource(DrbdRscData<Resource> drbdRscData, int timeout) throws ExtCmdFailedException
    {
        waitForFamily(drbdRscData, timeout, "wait-connect-resource");
    }

    public void waitSyncResource(DrbdRscData<Resource> drbdRscData, int timeout) throws ExtCmdFailedException
    {
        waitForFamily(drbdRscData, timeout, "wait-sync-resource");
    }

    public void checkResFile(
        Path tmpResPath,
        Path resPath
    )
        throws ExtCmdFailedException
    {
        String tmpResPathStr = tmpResPath.toString();
        execute(
            DRBDADM_UTIL, "--config-to-test", tmpResPathStr,
            "--config-to-exclude", resPath.toString(),
            "sh-nop"
        );

        // Using -c disables /etc/drbd.d/global_common.conf
        // execute(DRBDADM_UTIL, "-c", tmpResPathStr, "-d", "up", resourceName.value);
        // execute(DRBDADM_UTIL, "-d", "up", resourceName.value);
    }


    public void deletePeer(DrbdRscData<Resource> peerRscData) throws ExtCmdFailedException
    {
        deletePeer(peerRscData.getSuffixedResourceName(), peerRscData.getNodeId().value);
    }

    public void deletePeer(String suffixedRscName, int nodeId) throws ExtCmdFailedException
    {
        execute(
            DRBDSETUP_UTIL,
            "del-peer",
            suffixedRscName,
            Integer.toString(nodeId)
        );
    }

    public void forgetPeer(DrbdRscData<Resource> rscData) throws ExtCmdFailedException
    {
        forgetPeer(rscData.getSuffixedResourceName(), rscData.getNodeId().value);
    }

    public void forgetPeer(String suffixedRscName, int nodeId) throws ExtCmdFailedException
    {
        execute(
            DRBDSETUP_UTIL,
            "forget-peer",
            suffixedRscName,
            Integer.toString(nodeId)
        );
    }

    private void simpleSetupCommand(
        DrbdRscData<Resource> drbdRscData,
        VolumeNumber vlmNr,
        String... subCommands
    )
        throws ExtCmdFailedException
    {
        List<String> command = new ArrayList<>();
        command.add(DRBDSETUP_UTIL);
        command.addAll(Arrays.asList(subCommands));

        String drbdObj = drbdRscData.getSuffixedResourceName();
        if (vlmNr != null)
        {
            drbdObj += "/" + vlmNr.value;
        }
        command.add(drbdObj);

        execute(command);
    }

    private void simpleAdmCommand(DrbdRscData<Resource> drbdRscData, String subcommand) throws ExtCmdFailedException
    {
        simpleAdmCommand(drbdRscData, null, subcommand);
    }

    private void simpleAdmCommand(
        DrbdRscData<Resource> drbdRscData,
        VolumeNumber volNum,
        String... subCommands
    )
        throws ExtCmdFailedException
    {
        List<String> command = new ArrayList<>();
        command.add(DRBDADM_UTIL);
        command.add("-vvv");

        // Using -c disables /etc/drbd.d/global_common.conf
        // command.add("-c");
        // command.add("/etc/drbd.d/" + resourceName.displayValue + ".res");

        // command.addAll(asConfigParameter(resourceName.value));
        command.addAll(Arrays.asList(subCommands));
        String resName = drbdRscData.getSuffixedResourceName();
        if (volNum != null)
        {
            resName += "/" + volNum.value;
        }
        command.add(resName);

        execute(command);
    }

    private void waitForFamily(
        DrbdRscData<Resource> drbdRscData,
        int timeout,
        String commandRef
    )
        throws ExtCmdFailedException
    {
        execute(
            DRBDSETUP_UTIL,
            commandRef,
            "--wait-after-sb=yes",
            "--wfc-timeout=" + timeout,
            drbdRscData.getSuffixedResourceName()
        );
    }

    // Using -c disables /etc/drbd.d/global_common.conf
    // private List<String> asConfigParameter(String resourceName)
    // {
    //    String[] ret;
    //    if (!resourceName.toLowerCase().equals(ALL_KEYWORD) &&
    //        !resourceName.equals(DRBDCTRL_RES_NAME))
    //    {
    //       String resourcePath = configPath.resolve("linstor_" + resourceName + ".res").toString();
    //       ret = new String[]
    //       {
    //           "-c", resourcePath
    //       };
    //    }
    //    else
    //    {
    //       ret = new String[0];
    //    }
    //    return Arrays.asList(ret);
    // }

    private void execute(String... command) throws ExtCmdFailedException
    {
        execute(Arrays.asList(command));
    }

    private void execute(List<String> commandList) throws ExtCmdFailedException
    {
        String[] command = commandList.toArray(new String[commandList.size()]);
        try
        {
            // FIXME: Works only on Unix
            File nullDevice = new File("/dev/null");
            ExtCmd extCmd = extCmdFactory.create();
            OutputData outputData = extCmd.pipeExec(ProcessBuilder.Redirect.from(nullDevice), command);
            if (outputData.exitCode != 0)
            {
                throw new ExtCmdFailedException(command, outputData);
            }
        }
        catch (ChildProcessTimeoutException timeoutExc)
        {
            throw new ExtCmdFailedException(command, timeoutExc);
        }
        catch (IOException ioExc)
        {
            throw new ExtCmdFailedException(command, ioExc);
        }
    }
}
